"""This module defines the standard Lagrange finite element over
any shape for which an appropriate expansion basis is defined.  The
module simply provides the class Lagrange, which takes a shape and a
degree >= 1.

References for the Lagrange element are many.
Ciarlet: p. ???
Brenner & Scott: p. ???
"""

import dualbasis, functional, points, shapes, polynomial

class LagrangeDual( dualbasis.DualBasis ):
    """Dual basis for the plain vanilla Lagrange finite element."""
    def __init__( self , shape , n , U ):
        # Nodes are point evaluation on the lattice, which here is
        # ordered by topological dimension and entity within that
        # dimension.  For triangles we have vertex points plus
        # n - 1 points per edge.  We then have max(0,(n-2)(n-1)/2)
        # points on the interior of the triangle, or that many points
        # per face of tetrahedra plus max(0,(n-2)(n-1)n/6 points
        # in the interior
        ls = [ functional.PointEvaluation( U , pt ) \
               for d in shapes.dimension_range( shape ) \
               for v in shapes.entity_range( shape , d ) \
               for pt in points.make_points( shape , \
                                             d , v , n ) ] 

        # entity_ids is a dictionary whose keys are the topological
        # dimensions (0,1,2,3) and values are dictionaries mapping
        # the ids of entities of that dimension to the list of
        # ids of nodes associated with that entity.
        # see FIAT.base.dualbasis for description of entity_ids
        entity_ids = {}
        id_cur = 0
        for d in shapes.dimension_range( shape ):
            entity_ids[d] = {}
            for v in shapes.entity_range( shape, d ):
                num_nods = len( points.make_points( shape , \
                                                    d,v,n ) )
                entity_ids[d][v] = range(id_cur,id_cur + num_nods)
                id_cur += num_nods

        dualbasis.DualBasis.__init__( self , ls , entity_ids )
        return

class Lagrange( polynomial.FiniteElement ):
    def __init__( self , shape , n ):
        U = polynomial.OrthogonalPolynomialSet( shape , n )
        Udual = LagrangeDual( shape , n , U )
        polynomial.FiniteElement.__init__( self , \
                                           Udual , U )


